/*
    AVR-Tetris: A tetris game, running on an ATmega8, using
                an oscilloscope as a display.
    Copyright (C) 2010 Florian Jung (flo@windfisch.org)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// written for an ATmega8. Set the fuses to hfuse:lfuse = d9:e4.

#include <stdlib.h>
#include <stdio.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <util/delay.h>


#define INPUT_MOVE_LEFT  (1<<2)
#define INPUT_MOVE_RIGHT (1<<1)
#define INPUT_ROTATE_CW  (1<<4)
#define INPUT_ROTATE_CCW (1<<3)
#define INPUT_DROP       (1<<0)
#define INPUT_PLAY       (1<<0)
#define INPUT_TOGGLE_GRAVITY (1<<1)
#define INPUT_INC_SPEED (1<<2)
#define INPUT_DEC_SPEED (1<<3)
#define INPUT_OKAY      (1<<0)


#define DEBOUNCE 2

#define SCALE 3

#define DROP_SPEED 24

#define NEXT_BRICK_X 160
#define NEXT_BRICK_Y 192

#define SCORE_X	160
#define SCORE_Y 128

#define DRAW_DELAY 4
#define DRAW_DELAY_BHAM 2

#define CW  0
#define CCW 1

#define BRICK_XLEN 14
#define BRICK_YLEN 14
#define INSET 4
/* GRID_XLEN must NOT be greater than 12! */
#define GRID_XLEN 10
#define GRID_XBEGIN 3
#define GRID_XEND (GRID_XLEN+GRID_XBEGIN)
#define GRID_YLEN 20

#define GRID_EMPTY ( 7 | 1<<GRID_XEND ) /* *** ..xlen.. * */
#define GRID_FULL  ( (1<<(GRID_XEND+1)) - 1 )

#define FLAG_ANIM (1<<7)

#define MAX_BRICKS 100


#define PORTX PORTB
#define PORTY PORTD
#define DDRX  DDRB
#define DDRY  DDRD
#define PINI  PINC
#define DDRI  DDRC



#define FONT_VISIBLE (1<<7)
#define font_move(x,y) ((((x)<<4)+(y)) & (~FONT_VISIBLE))
#define font_line(x,y) ((((x)<<4)+(y)) | FONT_VISIBLE)
#define font_x(val) (((val)&0xF0&(~FONT_VISIBLE))>>4)
#define font_y(val) ((val)&0x0F)

const uint8_t font[] PROGMEM = {
24, 29, 32, 38, 44, 49, 55, 61, 64, 71, 77, 81, 86, 92, 96, 100, 104, 108, 112, 118, 124, 129, 132, 135,
font_move(4, 4), font_line(4,12), font_line(0,12), font_line(0, 4), font_line(4, 4),
font_move(0, 8), font_line(4,12), font_line(4, 4), 
font_move(0,12), font_line(4,12), font_line(4, 8), font_line(0, 8), font_line(0, 4), font_line(4, 4), 
font_move(0,12), font_line(4,12), font_line(4, 4), font_line(0, 4), font_move(4, 8), font_line(0, 8), 
font_move(0,12), font_line(0, 8), font_line(4, 8), font_move(4,12), font_line(4, 4), 
font_move(4,12), font_line(0,12), font_line(0, 8), font_line(4, 8), font_line(4, 4), font_line(0, 4), 
font_move(4,12), font_line(0,12), font_line(0, 4), font_line(4, 4), font_line(4, 8), font_line(0, 8), 
font_move(0,12), font_line(4,12), font_line(4, 4), 
font_move(0, 8), font_line(0,12), font_line(4,12), font_line(4, 4), font_line(0, 4), font_line(0, 8), font_line(4, 8), 
font_move(0, 4), font_line(4, 4), font_line(4,12), font_line(0,12), font_line(0, 8), font_line(4, 8), 
font_move(4,12), font_line(0,12), font_line(0, 4), font_line(4, 4), 
font_move(0, 4), font_line(0,12), font_line(4,10), font_line(0, 8), font_line(4, 4), 
font_move(4,12), font_line(0,12), font_line(0, 4), font_line(4, 4), font_move(0, 8), font_line(4, 8), 
font_move(0, 4), font_line(0,12), font_line(4, 4), font_line(4,12), 
font_move(4,12), font_line(0, 4), font_move(0,12), font_line(4, 4), 
font_move(2, 4), font_line(2,12), font_move(0,12), font_line(4,12), 
font_move(0, 4), font_line(0,12), font_line(4,10), font_line(0, 8), 
font_move(0, 4), font_line(0,12), font_line(4, 8), font_line(0, 4), 
font_move(1,12), font_line(3,12), font_move(2,12), font_line(2, 4), font_move(1, 4), font_line(3, 4), 
font_move(4,12), font_line(0,10), font_line(0, 6), font_line(4, 4), font_line(4, 8), font_line(2, 8), 
font_move(0, 4), font_line(2,12), font_line(4, 4), font_move(1, 7), font_line(3, 7), 
font_move(0,12), font_line(2, 4), font_line(4,12), 
font_move(0,12), font_line(0, 4), font_line(4, 4), 
};

const uint8_t text_tetris[] PROGMEM={15, 12, 15, 11, 18, 5, 0xFF};
const uint8_t text_score[] PROGMEM={5, 10, 0, 11, 12, 0xFF};
const uint8_t text_next[] PROGMEM={13, 12, 14, 15, 0xFF};
const uint8_t text_grav[] PROGMEM={19, 11, 20, 21, 0xFF};
const uint8_t text_orig[] PROGMEM={0, 11, 18, 19, 0xFF};
const uint8_t text_real[] PROGMEM={11, 12, 20, 22, 0xFF};
const uint8_t text_speed[] PROGMEM={5, 16, 12, 12, 17, 0xFF};

typedef uint16_t grid_t; /* grid_t has to have at least GRID_XLEN+4 */
												 /* bits, otherwise it can't hold the grid  */

typedef struct
{
 uint8_t data[4];
 uint8_t gridx;   /* MSB is used for animation */
 uint8_t gridy;
} brick_t;


int nbricks=0;
uint16_t score=0;

brick_t brick[MAX_BRICKS],curr_brick,next_brick;
grid_t grid[GRID_YLEN]; /* grid[0] has to be ~0 !   */

const uint8_t bricktemplate[]  PROGMEM={ 0x0F,0x00,
																			0x06,0x06,
																			0x02,0x07,
																			0x07,0x01,
																			0x07,0x04,
																			0x03,0x06,
																			0x06,0x03 };


inline uint8_t brickblock(uint8_t data[], int8_t x, int8_t y);
inline uint8_t is_empty_brick(brick_t* b);
inline void copy_brick (brick_t* dest, brick_t* src);
inline void add_brick (brick_t* b);
inline void remove_brick (int n);
inline uint8_t brickcollides (brick_t* b);
inline void xor_grid (brick_t* b);


void draw_brick(brick_t* brick, int add_y);
void rotatebrick(brick_t* dest, brick_t* src, int dir);
void split_brick(brick_t* lower, brick_t* upper, brick_t* mid, int l);
void remove_line(int8_t l);
uint8_t calc_gravity(int8_t real);


inline void drawlib_set_xy(uint8_t x, uint8_t y);
void drawlib_draw_dxdy(int8_t dx, int8_t dy, uint8_t cnt);
void wait_frame(void);


int main();
void game(uint8_t falling_speed, uint8_t grav_type);

void draw_normal(int8_t add_y);
inline void draw_curr_brick(int8_t add_y);
void draw_blinking(uint8_t blink);

void clear_grid();

void init_curr_brick();
void init_next_brick();

inline int move_and_check(brick_t* b, int dir);
inline int rotate_and_check (brick_t* b, int dir);

uint8_t get_input();


inline int sgn(int x){
  return (x > 0) ? 1 : (x < 0) ? -1 : 0;
}


void gbham(uint8_t xstart,uint8_t ystart,uint8_t xend,uint8_t yend) /* TODO stolen from german wikipedia */
/*--------------------------------------------------------------
 * Bresenham-Algorithmus: Linien auf Rastergeräten zeichnen
 *
 * Eingabeparameter:
 *    int xstart, ystart        = Koordinaten des Startpunkts
 *    int xend, yend            = Koordinaten des Endpunkts
 *
 * Ausgabe:
 *    void SetPixel(int x, int y) setze ein Pixel in der Grafik
 *         (wird in dieser oder aehnlicher Form vorausgesetzt)
 *---------------------------------------------------------------
 */
{
    int16_t x, y, t, dx, dy, incx, incy, pdx, pdy, ddx, ddy, es, el, err;
 
/* Entfernung in beiden Dimensionen berechnen */
   dx = xend - xstart;
   dy = yend - ystart;
 
/* Vorzeichen des Inkrements bestimmen */
   incx = sgn(dx);
   incy = sgn(dy);
   if(dx<0) dx = -dx;
   if(dy<0) dy = -dy;
 
/* feststellen, welche Entfernung größer ist */
   if (dx>dy)
   {
      /* x ist schnelle Richtung */
      pdx=incx; pdy=0;    /* pd. ist Parallelschritt */
      ddx=incx; ddy=incy; /* dd. ist Diagonalschritt */
      es =dy;   el =dx;   /* Fehlerschritte schnell, langsam */
   } else
   {
      /* y ist schnelle Richtung */
      pdx=0;    pdy=incy; /* pd. ist Parallelschritt */
      ddx=incx; ddy=incy; /* dd. ist Diagonalschritt */
      es =dx;   el =dy;   /* Fehlerschritte schnell, langsam */
   }
 
/* Initialisierungen vor Schleifenbeginn */
   x = xstart;
   y = ystart;
   err = el/2;
   PORTX=x;
   PORTY=y;
       
/* Pixel berechnen */
   for(t=0; t<el; ++t) /* t zaehlt die Pixel, el ist auch Anzahl */
   {
      /* Aktualisierung Fehlerterm */
      err -= es; 
      if(err<0)
      {
          /* Fehlerterm wieder positiv (>=0) machen */
          err += el;
          /* Schritt in langsame Richtung, Diagonalschritt */
          x += ddx;
          y += ddy;
      } else
      {
          /* Schritt in schnelle Richtung, Parallelschritt */
          x += pdx;
          y += pdy;
      }
      
      PORTX=x;
      PORTY=y;
      _delay_us(DRAW_DELAY_BHAM);
   }
} /* gbham() */


void write(uint8_t x, uint8_t y, uint8_t size, uint8_t no)
{
  uint8_t begin=pgm_read_byte(font+no);
  uint8_t end=pgm_read_byte(font+no+1);
  
  uint8_t i;
  uint8_t xx,yy;
  uint8_t tmp;
  
  for(i=begin;i<end;i++)
  {
    tmp=pgm_read_byte(font+i);
    if (tmp&FONT_VISIBLE)
      gbham(xx*size+x,yy*size+y, font_x(tmp)*size+x, font_y(tmp)*size+y);
    xx=font_x(tmp); yy=font_y(tmp);
  }
}
void write_str (uint8_t x, uint8_t y, uint8_t size, uint8_t* ptr)
{
	uint8_t tmp;
	while ((tmp=pgm_read_byte(ptr))!=0xFF)
	{
		write(x,y,size,tmp);
		ptr++;
		x+=6*size;
	}
}

void number(uint8_t x, uint8_t y, uint8_t size, uint16_t number)
{
  write(x,y,size,number/10000); x+=6*size; number=number%10000;
  write(x,y,size,number/ 1000); x+=6*size; number=number% 1000;
  write(x,y,size,number/  100); x+=6*size; number=number%  100;
  write(x,y,size,number/   10); x+=6*size; number=number%   10;
  write(x,y,size,number/    1); x+=6*size; number=number%    1;
}
void number2(uint8_t x, uint8_t y, uint8_t size, uint8_t number)
{
  write(x,y,size,number/   10); x+=6*size; number=number%   10;
  write(x,y,size,number/    1); x+=6*size; number=number%    1;
}

inline void write_next(uint8_t x, uint8_t y)
{
	write_str(x,y,2,text_next);
}
inline void write_score(uint8_t x, uint8_t y)
{
	write_str(x,y,2,text_score);
}



inline uint8_t brickblock(uint8_t data[], int8_t x, int8_t y)
{
	return ((uint8_t)x<4 && (uint8_t)y<4)?(data[y]&(1<<x)):0;
}
inline uint8_t is_empty_brick(brick_t* b)
{
	return ((b->data[0] | b->data[1] | b->data[2] | b->data[3]) & 0x0F)==0;
}
inline void copy_brick (brick_t* dest, brick_t* src)
{
	memcpy(dest,src,sizeof(brick_t));
}
inline void add_brick (brick_t* b)
{
	copy_brick (&brick[nbricks],b);
	nbricks++;
}
inline void remove_brick (int n)
{
	nbricks--;
	copy_brick (&brick[n],&brick[nbricks]);
}

/* assumes that brick is inside the range of the grid */
inline uint8_t brickcollides (brick_t* b)
{
	uint8_t tmp=b->gridx & 0x0F;
	return (
		( grid[b->gridy+0] & ( (((grid_t)b->data[0])&0x0F) << tmp ) ) ||
		( grid[b->gridy+1] & ( (((grid_t)b->data[1])&0x0F) << tmp ) ) ||
		( grid[b->gridy+2] & ( (((grid_t)b->data[2])&0x0F) << tmp ) ) ||
		( grid[b->gridy+3] & ( (((grid_t)b->data[3])&0x0F) << tmp ) ) );
}

/* assumes that brick is inside the range of the grid */
inline void xor_grid (brick_t* b)
{
	uint8_t tmp=b->gridx & 0x0F;
	grid[b->gridy+0] ^= ( (((grid_t)b->data[0])&0x0F) << tmp );
	grid[b->gridy+1] ^= ( (((grid_t)b->data[1])&0x0F) << tmp );
	grid[b->gridy+2] ^= ( (((grid_t)b->data[2])&0x0F) << tmp );
	grid[b->gridy+3] ^= ( (((grid_t)b->data[3])&0x0F) << tmp );
}


void draw_brick(brick_t* brick, int add_y) /* add_y=0xFF for next brick */
{
	int x,y,dx,dy,tmp,x0;

	x0=0; /* find the first block in our brick */
	while((brick->data[0]&(1<<x0))==0) x0++;

	if (add_y!=0xFF)
		drawlib_set_xy (((brick->gridx & 0x0F)-GRID_XBEGIN)*BRICK_XLEN + x0*BRICK_XLEN+4  ,   (brick->gridy-1)*BRICK_YLEN + ((brick->gridx&FLAG_ANIM)?add_y:0) + 4 );
	else
		drawlib_set_xy(NEXT_BRICK_X+x0*BRICK_XLEN,NEXT_BRICK_Y);

  x=x0; y=0; dx=1; dy=0;
  do {
		if (brickblock(brick->data, x+dy, y-dx)) {
      tmp = -dx;
      dx = dy;
      dy = tmp;

      drawlib_draw_dxdy(dx,dy, INSET);

      x += dx;
      y += dy;
    }
		else if (brickblock(brick->data, x+dx, y+dy)) {
      drawlib_draw_dxdy(dx,dy, BRICK_XLEN);

      x += dx;
      y += dy;
    }
    else {
      drawlib_draw_dxdy(dx,dy, BRICK_XLEN-INSET);

      tmp = dx;
      dx = -dy;
      dy = tmp;
    }
  } while (!((dx == 1) && (dy == 0) && (y == 0) && (x == x0)) );
}


/* do NOT apply on empty bricks! */
void rotatebrick(brick_t* dest, brick_t* src, int dir) /*0:CCW,1:CW*/
{
	int8_t x,y,move;
	uint8_t tmp[4];


	x=0; /* x is the begin of the brick */
	while (src->data[x]==0) x++;

	y=3; /* y is the end of the brick */
	while (src->data[y]==0) y--;

	move=(3-y-x+dir)>>1; /* move=diff/2, but with correct rounding: */
											/* if dir=0 round down, if dir=1 round up  */
	if (move>0) /* center brick in y (which will be x after rotating) */
	{           /* src: src, dest: tmp */
		for (y=3;y>=move;y--)
			tmp[y]=src->data[y-move];
		for (y=0;y<move;y++)
			tmp[y]=0;
	}
	else
	{  /* data[0]=data[0+move] i+move<4; i<4-move */
		for (y=0;y<4-move;y++)
			tmp[y]=src->data[y+move];
		for (y=4-move;y<4;y++)
			tmp[y]=0;
	}

	for (y=0;y<4;y++) dest->data[y]=0; /* clear dest */

	for (x=0;x<4;x++)     /* rotate the brick */
		for (y=0;y<4;y++)   /* src: tmp, dest: dest */
			if (brickblock(tmp,x,y))
				if (dir==CCW) /* CCW */
					dest->data[x]|=1<<(3-y);
				else        /* CW  */
					dest->data[3-x]|=1<<y;

	while(dest->data[0]==0) /* make the brick start at y=0 */
	{                        /* src: dest, dest: dest*/
		dest->data[0]=dest->data[1];
		dest->data[1]=dest->data[2];
		dest->data[2]=dest->data[3];
		dest->data[3]=0;
	}

	dest->gridx=src->gridx; /* anim flag is also copied */
	dest->gridy=src->gridy;
}

void split_brick(brick_t* lower, brick_t* upper, brick_t* mid, int l)
{
	int8_t i;

	copy_brick(lower,mid);
	copy_brick(upper,mid);

	for (i=l;i<4;i++)
		lower->data[i]=0;

	mid->data[0]=mid->data[l];
	for(i=1;i<4;i++)
		mid->data[i]=0;
	mid->gridy+=l;

	for (i=l+1;i<4;i++)
		upper->data[i-l-1]=upper->data[i];
	for (i=3-l;i<4;i++)
		upper->data[i]=0;
	upper->gridy+=l+1;
}

/* iterate through brick[] and split every brick that is in the line. */
/* then fadeout/blink/etc. the removed line and remove it             */
void remove_line(int8_t l)
{
	int8_t i,tmp;
	brick_t lower,upper;

	for (i=0;i<nbricks;i++)    /* split the bricks */
	{
		tmp=l-brick[i].gridy;
		if ((unsigned int)tmp<4) /* this brick is to split! */
			if (brick[i].data[tmp])  /* yep. */
			{
				split_brick(&lower,&upper,&brick[i],tmp);
				if (!is_empty_brick(&lower))  /* if it isn't empty: */
					add_brick(&lower);         /* add      */
				if (!is_empty_brick(&upper))
					add_brick(&upper);
				brick[i].gridx|=FLAG_ANIM; /* set the anim flag */
			}
	}

	for (i=0;i<16*4;i++)
	{
		wait_frame();
		draw_blinking((i%16)<8);
  }

	for (i=0;i<nbricks;i++)
		if (brick[i].gridy==l)
			remove_brick(i--);  /* remove it and decrease i for rechecking */

	grid[l]=GRID_EMPTY;  /* remove the whole line from the grid */
}

uint8_t calc_gravity(int8_t real)
{
	int8_t i,y,proc;

  for (i=0;i<nbricks;i++) brick[i].gridx&= ~FLAG_ANIM; /* clear anim */
	
	if (!real)
	{
		for (y=1;y<GRID_YLEN;y++)       /* line 0 must not be cleared */
			if (grid[y]==GRID_EMPTY)      /* search for empty lines */
			{
				proc=0;
				for (i=0;i<nbricks;i++)
					if (brick[i].gridy>y)     /* this brick has to be moved down */
						{
							xor_grid(&brick[i]);  /* remove from grid, */
							brick[i].gridy--;     /* move              */
							xor_grid(&brick[i]);  /* and add again     */
							brick[i].gridx|=FLAG_ANIM; /* this brick has to be animated */
							proc=1;
						}
				
				return proc; 								/* stop searching and return */
			}
		return 0;
	}
	else
	{
		proc=0;
		do
		{
			y=0;
			for (i=0;i<nbricks;i++)          /* check all bricks: */
				if (!(brick[i].gridx & FLAG_ANIM)) /* not already animated: */
				{
					xor_grid(&brick[i]);           /* temporarily remove from grid */
					brick[i].gridy--;              /* try moving down */
					if (!brickcollides(&brick[i])) /* moving down is ok. */
					{
						xor_grid(&brick[i]);         /* add at the new position */
						brick[i].gridx|=FLAG_ANIM;   /* and mark it as animated */
						proc=1; y=1;
					}
					else                           /* moving down is not ok */
					{
						brick[i].gridy++;            /* move up to the old position again */
						xor_grid(&brick[i]);         /* and add it there to the grid again */
					}
				}
		} while (y);												 /* changed something? then recheck! */
		return proc;
	}
}


inline void drawlib_set_xy(uint8_t x, uint8_t y)
{
  PORTX=x; PORTY=y;
}

void drawlib_draw_dxdy(int8_t dx, int8_t dy, uint8_t cnt)
{
  while(cnt--)
  {
    PORTX+=dx;
    PORTY+=dy;
    _delay_us(DRAW_DELAY);
  }
}

void wait_frame(void)
{
  uint16_t t=0;
  while((TIFR & (1<<OCF1A)) == 0)	/* while idle, draw the playing  */
  {																/* field to keep the beam moving */
    t++;
    if (t>=148+256+148+256) t=0;
    if (t<148)  drawlib_set_xy(t,0);
    else if (t<148+256)  drawlib_set_xy(148,t-148);
    else if (t<148+256+148)  drawlib_set_xy(148+256+148-t,255);
    else drawlib_set_xy(0,148+256+148+256-t);
  }
  TIFR = (1<<OCF1A); // write logic one to clear OCF1A
}



int main()
{
  uint8_t tmp;
	uint8_t settings=3 | (1<<7);
	
  DDRX=0xFF;
	DDRY=0xFF;
	DDRI=0x00;
  
	TCCR1B=(1<<WGM12) | (1<<CS10) | (1<<CS11);
	OCR1A=512*5; // 2048
  
  while (1)
	{
		tmp=get_input();
		if (tmp & INPUT_TOGGLE_GRAVITY)
		{
			if (settings & (1<<7))
				settings&= ~(1<<7);
			else
				settings|= (1<<7);
		}
		if (tmp & INPUT_INC_SPEED)
			if ( (settings&(~(1<<7))) < DROP_SPEED)
				settings++;
		
		if (tmp & INPUT_DEC_SPEED)
			if ( (settings&(~(1<<7))) > 1)
				settings--;
				
		if (tmp & INPUT_PLAY)
		{
			srand(TCNT1);
			game((settings&(~(1<<7))), settings&(1<<7));
			while (!(get_input()&INPUT_OKAY))
			{
				write_str(35,130,4,text_score);
				number(35,70,4,score);
			}
		}
		
		write_str(5,170,5,text_tetris);

		write_str(0,100,4,text_speed);
		number2( 150,110,3, (settings&(~(1<<7))));

		write_str(0,40,4,text_grav);
		if (settings&(1<<7))
			write_str(120,50,3,text_real);
		else
			write_str(120,50,3,text_orig);
	}
}








void game(uint8_t falling_speed, uint8_t grav_type)
{
	uint8_t lines_removed, linecount;
	uint8_t speed;
	uint8_t i;
	int8_t add_y;

	clear_grid();

	init_next_brick();
	init_curr_brick();

	init_next_brick();

	score=0;
	nbricks=0;
	speed=falling_speed;
	add_y=0;

	while(1)
	{
		switch (get_input())                       /* check input keys */
		{
			case INPUT_DROP:											   /* brick has been dropped? */
				if (speed!=DROP_SPEED) score+=(curr_brick.gridy-1);
				speed=DROP_SPEED;
				break;
			case INPUT_MOVE_LEFT:                    /* move left or right? */
				move_and_check(&curr_brick,-1);
				break;
			case INPUT_MOVE_RIGHT:
				move_and_check(&curr_brick, 1);
				break;
			case INPUT_ROTATE_CW:									   /* rotate the brick? */
				rotate_and_check(&curr_brick, CW);
				break;
			case INPUT_ROTATE_CCW:
				rotate_and_check(&curr_brick, CCW);
				break;
		}

		add_y-=speed;                              /* move brick down */ 
		if (add_y<0)                               /* new position in grid */
		{
			add_y+=BRICK_YLEN << SCALE;              /* add_y has to stay in range */
			curr_brick.gridy--;

			if (brickcollides(&curr_brick))          /* would the brick collide? */
			{
				curr_brick.gridx&= ~FLAG_ANIM;         /* reset position to the */
				curr_brick.gridy++;                    /* last valid position   */

				add_brick(&curr_brick);
				xor_grid(&curr_brick);                 /* add it to the grid */
				
				linecount=0;
				do
				{
					lines_removed=0;
					for (i=0;i<GRID_YLEN;i++)            /* search for full lines */
						if (grid[i]==GRID_FULL)            /* full line? */
						{
							remove_line(i);									 /* remove that line */
							lines_removed++;
							linecount++;
						}

					if (lines_removed)                   /* if there was at least one */
					{                                    /* full line:                */
							add_y=BRICK_YLEN << SCALE;
							while (calc_gravity(grav_type))  /* still falling bricks? */
							{
								while(1)
								{
									add_y-=DROP_SPEED;
									if (add_y<0) 								 /* bricks have fallen */
										break; 										 /* one cube: break;   */

									wait_frame();
									draw_normal(add_y >> SCALE);
								}
								add_y+=BRICK_YLEN << SCALE;
							}
		  		}
				} while (lines_removed);               /* and check again */
				
				score+=(linecount*linecount)*10;			 /* scorekeeping */
				
				add_y=0;
				init_curr_brick();                     /* get a new brick to play */
				if (brickcollides(&curr_brick))        /* if it can't be placed, */
					break;                               /* the game is over       */

				init_next_brick();                     /* and get a new brick */
                                               /* as next brick       */
				speed=falling_speed;									 /* and reset speed to normal */
      }
		}
    wait_frame();
		draw_curr_brick(add_y >> SCALE);
    draw_normal(add_y >> SCALE);

    drawlib_set_xy(255,255);
	}
}

void draw_normal(int8_t add_y)
{
  uint8_t i;
  for (i=0;i<nbricks;i++)
    draw_brick(&brick[i], add_y);
    
  draw_brick(&next_brick, 0xFF);
	
	write_next(NEXT_BRICK_X, NEXT_BRICK_Y+2*BRICK_YLEN);
	write_score(SCORE_X,SCORE_Y);
	number(SCORE_X,SCORE_Y-12*2,2,score);
}

inline void draw_curr_brick(int8_t add_y)
{
	draw_brick(&curr_brick, add_y);draw_brick(&curr_brick, add_y);draw_brick(&curr_brick, add_y);
}

void draw_blinking(uint8_t blink)
{
  uint8_t i;
  for (i=0;i<nbricks;i++)
    if ((!(brick[i].gridx&FLAG_ANIM))||blink)
			draw_brick(&brick[i], 0);
    
  draw_brick(&next_brick, 0xFF);

	write_next(NEXT_BRICK_X, NEXT_BRICK_Y+2*BRICK_YLEN);
	write_score(SCORE_X,SCORE_Y);
	number(SCORE_X,SCORE_Y-12*2,2,score);
}

void clear_grid()
{
	int8_t i;
	for (i=1;i<GRID_YLEN;i++)
		grid[i]=GRID_EMPTY;
	grid[0]=~0;
}

void init_curr_brick()
{
	copy_brick(&curr_brick,&next_brick);
	curr_brick.gridx=(GRID_XEND-GRID_XBEGIN)/2-2;
	curr_brick.gridy=GRID_YLEN-5;
	curr_brick.gridx|=FLAG_ANIM;
}

void init_next_brick()
{
	uint8_t* offset=bricktemplate + (rand()%7)*2;
	
	next_brick.data[0]=pgm_read_byte(offset);
	next_brick.data[1]=pgm_read_byte(offset+1);
	next_brick.data[2]=0;
	next_brick.data[3]=0;
}

inline int move_and_check(brick_t* b, int dir) /* returns 1 on success */
{
	b->gridx+=dir;
	if  (brickcollides(b))
	{
		b->gridx-=dir;
		return 0;
	}
	else
	{
		return 1;
	}
}

inline int rotate_and_check (brick_t* b, int dir)
{
	brick_t temp;

	rotatebrick(&temp,b,dir);
	if (brickcollides(&temp))
	{
		if (!move_and_check(&temp,-1))
			if (!move_and_check(&temp,1))
				if (!move_and_check(&temp,-2)) /* fail */
					return 0;
	}
	/* success? */
	copy_brick(b,&temp);
	return 1;
}


uint8_t get_input()
{
  static uint8_t count[8];
  
  uint8_t tmp=PINI;
  uint8_t result=0;
  uint8_t i;
  for (i=0;i<8;i++)       /* for each switch: */
  {
    if (!(tmp&(1<<i)))    /* pressed? */
    {
      if (count[i]==0)    /* count is zero? */
        result|=(1<<i);   /* report keypress */
      count[i]=DEBOUNCE;  /* set debounce counter */
    }
    else                  /* not pressed? */
    {
      if (count[i])       /* if counter>0 */
        count[i]--;       /* decrease counter */
    }
  }
  
  return result;
}
